@@ -2,13 +2,13 @@
 package Module::Install::PAR;
 
 use strict;
-use Module::Install::Base;
+use Module::Install::Base ();
 
-use vars qw{$VERSION $ISCORE @ISA};
+use vars qw{$VERSION @ISA $ISCORE};
 BEGIN {
-	$VERSION = '0.80';
+	$VERSION = '1.17';
+	@ISA     = 'Module::Install::Base';
 	$ISCORE  = 1;
-	@ISA     = qw{Module::Install::Base};
 }
 
 #line 74
@@ -117,7 +117,7 @@ sub extract_par {
     else {
         die <<'HERE';
 Could not extract .par archive because neither Archive::Zip nor a
-working 'unzip' binary are availlable. Please consider installing
+working 'unzip' binary are available. Please consider installing
 Archive::Zip.
 HERE
     }
@@ -135,7 +135,7 @@ sub make_par {
     my ($self, $file) = @_;
     unlink $file if -f $file;
 
-    unless ( eval { require PAR::Dist; PAR::Dist->VERSION >= 0.03 } ) {
+    unless ( eval { require PAR::Dist; PAR::Dist->VERSION(0.03) } ) {
         warn "Please install PAR::Dist 0.03 or above first.";
         return;
     }
