@@ -1,5 +1,4 @@
 #!/usr/bin/perl
-
 eval 'exec /usr/bin/perl  -S $0 ${1+"$@"}'
     if 0; # not running under some shell
 
@@ -242,14 +241,29 @@ my ($start_pos, $data_pos);
     open _FH, '<', $progname or last;
     binmode(_FH);
 
+    # Search for the "\nPAR.pm\n signature backward from the end of the file
     my $buf;
-    seek _FH, -8, 2;
-    read _FH, $buf, 8;
-    last unless $buf eq "\nPAR.pm\n";
+    my $size = -s $progname;
+    my $offset = 512;
+    my $idx = -1;
+    while (1)
+    {
+        $offset = $size if $offset > $size;
+        seek _FH, -$offset, 2 or die qq[seek failed on "$progname": $!];
+        my $nread = read _FH, $buf, $offset;
+        die qq[read failed on "$progname": $!] unless $nread == $offset;
+        $idx = rindex($buf, "\nPAR.pm\n");
+        last if $idx >= 0 || $offset == $size || $offset > 128 * 1024;
+        $offset *= 2;
+    }
+    last unless $idx >= 0;
 
-    seek _FH, -12, 2;
+    # Seek 4 bytes backward from the signature to get the offset of the 
+    # first embedded FILE, then seek to it
+    $offset -= $idx - 4;
+    seek _FH, -$offset, 2;
     read _FH, $buf, 4;
-    seek _FH, -12 - unpack("N", $buf), 2;
+    seek _FH, -$offset - unpack("N", $buf), 2;
     read _FH, $buf, 4;
 
     $data_pos = (tell _FH) - 4;
@@ -270,24 +284,13 @@ my ($start_pos, $data_pos);
         read _FH, $buf, unpack("N", $buf);
 
         if (defined($ext) and $ext !~ /\.(?:pm|pl|ix|al)$/i) {
-            my ($out, $filename) = _tempfile($ext, $crc);
-            if ($out) {
-                binmode($out);
-                print $out $buf;
-                close $out;
-                chmod 0755, $filename;
-            }
+            my $filename = _tempfile("$crc$ext", $buf, 0755);
             $PAR::Heavy::FullCache{$fullname} = $filename;
             $PAR::Heavy::FullCache{$filename} = $fullname;
         }
         elsif ( $fullname =~ m|^/?shlib/| and defined $ENV{PAR_TEMP} ) {
-            # should be moved to _tempfile()
-            my $filename = "$ENV{PAR_TEMP}/$basename$ext";
+            my $filename = _tempfile("$basename$ext", $buf, 0755);
             outs("SHLIB: $filename\n");
-            open my $out, '>', $filename or die $!;
-            binmode($out);
-            print $out $buf;
-            close $out;
         }
         else {
             $require_list{$fullname} =
@@ -325,13 +328,9 @@ my ($start_pos, $data_pos);
             return $fh;
         }
         else {
-            my ($out, $name) = _tempfile('.pm', $filename->{crc});
-            if ($out) {
-                binmode($out);
-                print $out $filename->{buf};
-                close $out;
-            }
-            open my $fh, '<', $name or die $!;
+            my $filename = _tempfile("$filename->{crc}.pm", $filename->{buf});
+
+            open my $fh, '<', $filename or die "can't read $filename: $!";
             binmode($fh);
             return $fh;
         }
@@ -378,8 +377,9 @@ my ($start_pos, $data_pos);
 my @par_args;
 my ($out, $bundle, $logfh, $cache_name);
 
-$quiet = 0 unless $ENV{PAR_DEBUG};
+delete $ENV{PAR_APP_REUSE}; # sanitize (REUSE may be a security problem)
 
+$quiet = 0 unless $ENV{PAR_DEBUG};
 # Don't swallow arguments for compiled executables without --par-options
 if (!$start_pos or ($ARGV[0] eq '--par-options' && shift)) {
     my %dist_cmd = qw(
@@ -390,51 +390,61 @@ if (!$start_pos or ($ARGV[0] eq '--par-options' && shift)) {
         v   verify_par
     );
 
-    my @add_to_inc;
-    while (@ARGV) {
-        $ARGV[0] =~ /^-([AIMOBLbqpiusTv])(.*)/ or last;
-
-        if ($1 eq 'I') {
-            push @add_to_inc, $2;
-        }
-        elsif ($1 eq 'M') {
-            eval "use $2";
-        }
-        elsif ($1 eq 'A') {
-            unshift @par_args, $2;
-        }
-        elsif ($1 eq 'O') {
-            $out = $2;
-        }
-        elsif ($1 eq 'b') {
-            $bundle = 'site';
-        }
-        elsif ($1 eq 'B') {
-            $bundle = 'all';
-        }
-        elsif ($1 eq 'q') {
-            $quiet = 1;
-        }
-        elsif ($1 eq 'L') {
-            open $logfh, ">>", $2 or die "XXX: Cannot open log: $!";
-        }
-        elsif ($1 eq 'T') {
-            $cache_name = $2;
-        }
-
-        shift(@ARGV);
-
-        if (my $cmd = $dist_cmd{$1}) {
-            delete $ENV{'PAR_TEMP'};
-            init_inc();
-            require PAR::Dist;
-            &{"PAR::Dist::$cmd"}() unless @ARGV;
-            &{"PAR::Dist::$cmd"}($_) for @ARGV;
-            exit;
-        }
+    # if the app is invoked as "appname --par-options --reuse PROGRAM @PROG_ARGV",
+    # use the app to run the given perl code instead of anything from the
+    # app itself (but still set up the normal app environment and @INC)
+    if (@ARGV and $ARGV[0] eq '--reuse') {
+        shift @ARGV;
+        $ENV{PAR_APP_REUSE} = shift @ARGV;
     }
+    else { # normal parl behaviour
 
-    unshift @INC, @add_to_inc;
+        my @add_to_inc;
+        while (@ARGV) {
+            $ARGV[0] =~ /^-([AIMOBLbqpiusTv])(.*)/ or last;
+
+            if ($1 eq 'I') {
+                push @add_to_inc, $2;
+            }
+            elsif ($1 eq 'M') {
+                eval "use $2";
+            }
+            elsif ($1 eq 'A') {
+                unshift @par_args, $2;
+            }
+            elsif ($1 eq 'O') {
+                $out = $2;
+            }
+            elsif ($1 eq 'b') {
+                $bundle = 'site';
+            }
+            elsif ($1 eq 'B') {
+                $bundle = 'all';
+            }
+            elsif ($1 eq 'q') {
+                $quiet = 1;
+            }
+            elsif ($1 eq 'L') {
+                open $logfh, ">>", $2 or die "XXX: Cannot open log: $!";
+            }
+            elsif ($1 eq 'T') {
+                $cache_name = $2;
+            }
+
+            shift(@ARGV);
+
+            if (my $cmd = $dist_cmd{$1}) {
+                delete $ENV{'PAR_TEMP'};
+                init_inc();
+                require PAR::Dist;
+                &{"PAR::Dist::$cmd"}() unless @ARGV;
+                &{"PAR::Dist::$cmd"}($_) for @ARGV;
+                exit;
+            }
+        }
+
+        unshift @INC, @add_to_inc;
+    }
 }
 
 # XXX -- add --par-debug support!
@@ -454,6 +464,10 @@ if ($out) {
 
 
     if (defined $par) {
+        # increase the chunk size for Archive::Zip so that it will find the EOCD
+        # even if more stuff has been appended to the .par
+        Archive::Zip::setChunkSize(128*1024);
+
         open my $fh, '<', $par or die "Cannot find '$par': $!";
         binmode($fh);
         bless($fh, 'IO::File');
@@ -518,26 +532,18 @@ if ($out) {
 
         require_modules();
 
-        my @inc = sort {
-            length($b) <=> length($a)
-        } grep {
-            !/BSDPAN/
-        } grep {
-            ($bundle ne 'site') or
-            ($_ ne $Config::Config{archlibexp} and
-             $_ ne $Config::Config{privlibexp});
-        } @INC;
+        my @inc = grep { !/BSDPAN/ } 
+                       grep {
+                           ($bundle ne 'site') or
+                           ($_ ne $Config::Config{archlibexp} and
+                           $_ ne $Config::Config{privlibexp});
+                       } @INC;
 
-        if ($Config{_delim} eq '\\') { s{\\}{/}g for @inc }
-
-        # File exists test added to fix RT #41790:
-        # Funny, non-existing entry in _<....auto/Compress/Raw/Zlib/autosplit.ix.
-        # This is a band-aid fix with no deeper grasp of the issue.
-        # Somebody please go through the pain of understanding what's happening,
-        # I failed. -- Steffen
+        # Now determine the files loaded above by require_modules():
+        # Perl source files are found in values %INC and DLLs are
+        # found in @DynaLoader::dl_shared_objects.
         my %files;
-        /^_<(.+)$/ and -e $1 and $files{$1}++ for keys %::;
-        $files{$_}++ for values %INC;
+        $files{$_}++ for @DynaLoader::dl_shared_objects, values %INC;
 
         my $lib_ext = $Config::Config{lib_ext};
         my %written;
@@ -586,9 +592,7 @@ if ($out) {
                 PAR::Filter::PodStrip->new->apply(\$content, $file)
                     if !$ENV{PAR_VERBATIM} and $name =~ /\.(?:pm|ix|al)$/i;
 
-                # Do not let XSLoader pick up auto/* from environment
-                $content =~ s/goto +retry +unless +.*/goto retry;/
-                    if lc($name) eq lc("XSLoader.pm");
+                PAR::Filter::PatchContent->new->apply(\$content, $file, $name);
             }
 
             outs(qq(Written as "$name"));
@@ -742,11 +746,15 @@ sub require_modules {
     require PAR::Heavy;
     require PAR::Dist;
     require PAR::Filter::PodStrip;
+    require PAR::Filter::PatchContent;
+    require attributes;
     eval { require Cwd };
     eval { require Win32 };
     eval { require Scalar::Util };
     eval { require Archive::Unzip::Burst };
     eval { require Tie::Hash::NamedCapture };
+    eval { require PerlIO; require PerlIO::scalar };
+    eval { require utf8 };
 }
 
 # The C version of this code appears in myldr/mktmpdir.c
@@ -778,7 +786,7 @@ sub _set_par_temp {
         }
         $username =~ s/\W/_/g;
 
-        my $stmpdir = "$path$Config{_delim}par-$username";
+        my $stmpdir = "$path$Config{_delim}par-".unpack("H*", $username);
         mkdir $stmpdir, 0755;
         if (!$ENV{PAR_CLEAN} and my $mtime = (stat($progname))[9]) {
             open (my $fh, "<". $progname);
@@ -825,23 +833,33 @@ sub _set_par_temp {
     $par_temp = $1 if $ENV{PAR_TEMP} and $ENV{PAR_TEMP} =~ /(.+)/;
 }
 
+
+# check if $name (relative to $par_temp) already exists;
+# if not, create a file with a unique temporary name, 
+# fill it with $contents, set its file mode to $mode if present;
+# finaly rename it to $name; 
+# in any case return the absolute filename
 sub _tempfile {
-    my ($ext, $crc) = @_;
-    my ($fh, $filename);
+    my ($name, $contents, $mode) = @_;
 
-    $filename = "$par_temp/$crc$ext";
+    my $fullname = "$par_temp/$name";
+    unless (-e $fullname) {
+        my $tempname = "$fullname.$$";
 
-    if ($ENV{PAR_CLEAN}) {
-        unlink $filename if -e $filename;
-        push @tmpfile, $filename;
-    }
-    else {
-        return (undef, $filename) if (-r $filename);
+        open my $fh, '>', $tempname or die "can't write $tempname: $!";
+        binmode $fh;
+        print $fh $contents;
+        close $fh;
+        chmod $mode, $tempname if defined $mode;
+
+        rename($tempname, $fullname) or unlink($tempname);
+        # NOTE: The rename() error presumably is something like ETXTBSY 
+        # (scenario: another process was faster at extraction $fullname
+        # than us and is already using it in some way); anyway, 
+        # let's assume $fullname is "good" and clean up our copy.
     }
 
-    open $fh, '>', $filename or die $!;
-    binmode($fh);
-    return($fh, $filename);
+    return $fullname;
 }
 
 # same code lives in PAR::SetupProgname::set_progname
@@ -969,11 +987,11 @@ L<PAR>, L<PAR::Dist>, L<parl>, L<pp>
 Audrey Tang E<lt>cpan@audreyt.orgE<gt>,
 Steffen Mueller E<lt>smueller@cpan.orgE<gt>
 
-L<http://par.perl.org/> is the official PAR website.  You can write
+You can write
 to the mailing list at E<lt>par@perl.orgE<gt>, or send an empty mail to
 E<lt>par-subscribe@perl.orgE<gt> to participate in the discussion.
 
-Please submit bug reports to E<lt>bug-par@rt.cpan.orgE<gt>.
+Please submit bug reports to E<lt>bug-par-packer@rt.cpan.orgE<gt>.
 
 =head1 COPYRIGHT
 
@@ -996,7 +1014,7 @@ under the Artistic License.
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
 
-See L<http://www.perl.com/perl/misc/Artistic.html>
+See F<LICENSE>.
 
 =cut
 
